unit main;

(*******************************************************************************
Application:     PDFTK Builder
Description:     A graphical interface to the Windows version of PDFTK.
Module:          Main
Version:         3.6
Date:            9-JUL-2009
Compiler:        Delphi 7
Author:          Angus Johnson <angusj@myrealbox.com>
                 http://www.angusj.com
Copyright &       2005-09 Angus Johnson
License:         GNU GPL Ver 2 (see http://www.gnu.org/copyleft/gpl.html)
*******************************************************************************)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, IniFiles, ShlObj, ShellApi, ActiveX, ComCtrls, Grids, Menus, StrUtils,
  ExtCtrls, zlib;

type
  TMainForm = class(TForm)
    gbAdvanced: TGroupBox;
    GroupBox4: TGroupBox;
    btnBuild: TButton;
    bClose: TButton;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    cbAllowAll: TCheckBox;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Close1: TMenuItem;
    Help1: TMenuItem;
    StatusBar1: TStatusBar;
    Label4: TLabel;
    PageControl1: TPageControl;
    tsCollate: TTabSheet;
    Label2: TLabel;
    lFilename: TLabel;
    lPages: TLabel;
    bAdd: TButton;
    bDelete: TButton;
    bMoveUp: TButton;
    bMoveDown: TButton;
    StringGrid1: TStringGrid;
    ePages: TEdit;
    tsBackground: TTabSheet;
    btnSrcBackground: TButton;
    Button2: TButton;
    lblSource_background: TLabel;
    lblBackground: TLabel;
    About1: TMenuItem;
    Contents1: TMenuItem;
    tsSplit: TTabSheet;
    btnSplitSrc: TButton;
    lblSplitSrc: TLabel;
    btnSort: TButton;
    cbPageOneOnly: TCheckBox;
    GroupBox1: TGroupBox;
    rbBackground: TRadioButton;
    rbStamp: TRadioButton;
    tsRotate: TTabSheet;
    Label8: TLabel;
    eRotPagesFrom: TEdit;
    lblSource_rotate: TLabel;
    btnSrcRotate: TButton;
    GroupBox2: TGroupBox;
    rbRotLeft: TRadioButton;
    rbRotRight: TRadioButton;
    rbRot180: TRadioButton;
    Label6: TLabel;
    eRotPagesTo: TEdit;
    Panel1: TPanel;
    cbPrinting: TCheckBox;
    cbDegradedPrinting: TCheckBox;
    cbModifyContents: TCheckBox;
    cbScreenReaders: TCheckBox;
    cbModifyAnnotations: TCheckBox;
    cbFillIn: TCheckBox;
    cbAssembly: TCheckBox;
    cbCopyContents: TCheckBox;
    GroupBox3: TGroupBox;
    Label1: TLabel;
    eOwnerPassword: TEdit;
    Label5: TLabel;
    eUserPassword: TEdit;
    procedure FormCreate(Sender: TObject);
    procedure bCloseClick(Sender: TObject);
    procedure btnBuildClick(Sender: TObject);
    procedure bAddClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure StringGrid1SelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure bMoveUpClick(Sender: TObject);
    procedure bMoveDownClick(Sender: TObject);
    procedure bDeleteClick(Sender: TObject);
    procedure ePagesChange(Sender: TObject);
    procedure cbAllowAllClick(Sender: TObject);
    procedure cbPrintingClick(Sender: TObject);
    procedure PageControl1Change(Sender: TObject);
    procedure btnSrcBackgroundClick(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure Contents1Click(Sender: TObject);
    procedure btnSplitSrcClick(Sender: TObject);
    procedure btnSortClick(Sender: TObject);
    procedure eRotPagesFromChange(Sender: TObject);
    procedure btnSrcRotateClick(Sender: TObject);
    procedure tsCollateResize(Sender: TObject);
  private
    fFileList: TStringList;
    fRotatePageCnt: integer;
    SgOldWindowProc: TWndMethod;
    procedure SgNewWindowProc(var Message: TMessage);
    procedure RefreshGrid(RowSelected: integer);
    procedure ClearFileList;
    procedure UpdateButtonsAndMenus(RowSelected: integer);
    function FmtPagesString(PagesStr: string; out IsValid: boolean): string;
    procedure DeleteAllTmpPdfFiles;
  public
    { Public declarations }
  end;

  PSourceFileRec = ^TSourceFileRec;
  TSourceFileRec = record
    FullPath: string;
    Filename: string;
    Pages: string;
    PageCount: integer;
  end;

var
  MainForm: TMainForm;

  Path, tmpPath: string;
  ErrorMessage: string;
  ErrorNum: DWORD;

const
  VersionInfo = '  Version 3.6  '; //see also VersionInfo in *.res
  PAGES_NOT_COUNTED = -2;

implementation

uses About;

{$R *.dfm}

const
  NEVER_GIVE_UP = 0; //GiveUpTimeOutSecs param constant for WinExecAndWait32()

//function GetPdfPageCount(filename: pAnsiChar): integer; stdCall;
//  external 'PdfPageCount' name 'GetPdfPageCount';

//static extern int GetPdfPageCount(char *filename);

function WinExecAndWait32(const DosCommand: string;
  ShowWindow, GiveUpTimeOutSecs: Word; out textOutput: string): DWord;
const
  BufferSize = 8192;
var
  StartUpInfo:TStartupInfo;
  ProcessInfo:TProcessInformation;
  OutputReadPipeHdl, OutputWritePipeHdl: THandle;
  SecAttribs: TSecurityAttributes;
  Buffer: PChar;
  timeCnt, BytesRead, BytesAvailable, WaitResult: DWord;
  PipeCreated, ProcessCreated: boolean;

begin
  //nb: 1. The max command line length for CreateProcess() is 32767 characters.

  Result := DWORD(-1); //ie result when unable to create process

  if GiveUpTimeOutSecs = NEVER_GIVE_UP then
    GiveUpTimeOutSecs := $FFFF; //(ie about 18hrs so not quite never :))

  SecAttribs.nLength := SizeOf(TSecurityAttributes) ;
  SecAttribs.bInheritHandle := true;
  SecAttribs.lpSecurityDescriptor := nil;

  PipeCreated := CreatePipe(OutputReadPipeHdl,
    OutputWritePipeHdl, @SecAttribs, BufferSize);

  StartUpInfo.cb := Sizeof(StartUpInfo);
  StartUpInfo.wShowWindow := ShowWindow;

  Buffer := AllocMem(BufferSize + 1);
  try
    if PipeCreated then
    begin
      fillChar( startUpInfo, sizeof(startUpInfo), 0);
      startUpInfo.dwFlags := STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
      startUpInfo.hStdInput := 0;
      startUpInfo.hStdOutput := OutputWritePipeHdl;
      startUpInfo.hStdError := OutputWritePipeHdl;
      ProcessCreated := CreateProcess(nil, PChar(DosCommand),
        @SecAttribs, @SecAttribs, true, NORMAL_PRIORITY_CLASS,
        nil, nil, startUpInfo, ProcessInfo);
      textOutput := '';
    end else
    begin
      startUpInfo.dwFlags := STARTF_USESHOWWINDOW;
      ProcessCreated := CreateProcess(nil, PChar(DosCommand),
        nil, nil, false, NORMAL_PRIORITY_CLASS or CREATE_NEW_CONSOLE,
        nil, nil, startUpInfo, ProcessInfo);
    end;

    if ProcessCreated then
    begin
      timeCnt := 0;
      repeat
        WaitResult := WaitForSingleObject(ProcessInfo.hProcess, 100) ;

        if PipeCreated then
        begin
          //nb: a full pipe buffer would cause an endless loop ...
          if not PeekNamedPipe(OutputReadPipeHdl,
            nil, 0, nil, @BytesAvailable, nil) then break;
          if (BytesAvailable > 0) then
          begin
            //Interestingly, it appears that if the default pipe buffer is larger
            //that that supplied in CreatePipe(), then the default size is used.
            BytesRead := 0;
            if BytesAvailable > BufferSize then
              ReadFile(OutputReadPipeHdl, Buffer[0], BufferSize, BytesRead, nil)
            else
              ReadFile(OutputReadPipeHdl, Buffer[0], BytesAvailable, BytesRead, nil);
            Buffer[BytesRead]:= #0;
            OemToAnsi(Buffer, Buffer) ;
            textOutput := textOutput + Buffer;
          end;
        end;

        Application.ProcessMessages;
        inc(timeCnt, 100);
      until (WaitResult <> WAIT_TIMEOUT) or (timeCnt > GiveUpTimeOutSecs *1000);

      if not GetExitCodeProcess(ProcessInfo.hProcess, Result)
        and (result = STILL_ACTIVE) then
          TerminateProcess(ProcessInfo.hProcess, result);

      CloseHandle( ProcessInfo.hProcess );
      CloseHandle( ProcessInfo.hThread );
    end;

    if PipeCreated then
    begin
      CloseHandle(OutputWritePipeHdl);
      CloseHandle(OutputReadPipeHdl);
    end;
  finally
    FreeMem(Buffer) ;
  end;
end;
//------------------------------------------------------------------------------

type
{$IFNDEF UNICODE}
  PByte = PAnsiChar;
{$ENDIF}

  PPdfObj = ^TPdfObj;
  TPdfObj = record
    number,
    offset: integer;
    filePtr: pchar;
    stmObjNum: integer;
  end;

function ReversePngFilter(var buffer: PByte; row_size, buffer_cnt: integer): integer;
var
  i: integer;
  p, end_buff: PByte;
  filterType: AnsiChar;

  function paeth(c, b, a: byte): byte;
  var
    p,pa,pb,pc: byte;
  begin
    //a = left, b = above, c = upper left
    p := (a + b - c) and $FF;
    pa := abs(p - a);
    pb := abs(p - b);
    pc := abs(p - c);
    if (pa <= pb) and (pa <= pc) then result := a
    else if pb <= pc then result := b
    else result := c;
  end;

  function Average(a, b: byte): byte;
  begin
    //a = left, b = above
    result := (a + b) shr 1;
  end;

begin
  //result will be new buffer length ...
  result := buffer_cnt;
  p := buffer;
  end_buff := buffer + buffer_cnt;
  while p < end_buff do
  begin
    filterType := p^;
    dec(result);
    dec(end_buff);
    move((p +1)^, p^, end_buff - p);
    case filterType of
      #0:
        ;//no filtering used for this row
      #1:
        begin
          //'sub' filtering
          for i := 1 to row_size -1 do
            (p+i)^ := chr((ord((p+i)^) + ord((p+i -1)^)) and $FF);
        end;
      #2:
        begin
          //'up' filtering
          if p > buffer then //leave top row alone
            for i := 0 to row_size -1 do
              (p+i)^ := chr((ord((p+i)^) + ord((p+i - row_size)^)) and $FF);
        end;
      #3:
        begin
          //'average' filtering
          if p = buffer then //ie top row
            for i := 1 to row_size -1 do
              (p+i)^ := chr((ord((p+i)^) + average(ord((p+i -1)^), 0)) and $FF)
          else
          begin
            p^ := chr((ord((p)^) + average(0, ord((p - row_size)^))) and $FF);
            for i := 1 to row_size -1 do
              (p+i)^ := chr((ord((p+i)^) +
                average(ord((p+i -1)^), ord((p+i - row_size)^))) and $FF);
          end;
        end;
      #4:
        begin
          //'paeth' filtering
          if p = buffer then //ie top row
            for i := 1 to row_size -1 do
              (p+i)^ := chr((ord((p+i)^) + paeth(ord((p+i -1)^), 0, 0)) and $FF)
          else
          begin
            p^ := chr((ord((p)^) + paeth(0, ord((p - row_size)^), 0)) and $FF);
            for i := 1 to row_size -1 do
              (p+i)^ := chr((ord((p+i)^) + paeth(ord((p+i -1)^),
                ord((p+i - row_size)^), ord((p+i - row_size -1)^))) and $FF);
          end;
        end;
    end;
    inc(p, row_size);
  end;
  ReallocMem(buffer, result);
end;
//------------------------------------------------------------------------------

function GetPdfPageCount(const filename: string): integer;
var
  ms: TMemoryStream;
  i, j, k, cnt, pagesNum, rootNum, predictor, filterColCnt: integer;
  indexArray: array of integer;
  w1,w2,w3: integer;
  p, pEnd, pSaved, pTmp2, buffer, buffPtr: PByte;
  PdfObj: PPdfObj;
  PdfObjList: TList;

////////////////////////////////////////////////////////////////////////////////
// Summary of steps taken to parse a PDF doc for its page count :-             
// (Who'd have thought it'd take about 1000 lines of code to do that!?) 
////////////////////////////////////////////////////////////////////////////////

//1.  See if there's a 'Linearization dictionary' for easy parsing.
//    Mostly there isn't so ...
//2.  Locate 'startxref' at end of file
//3.  get 'xref' offset and go to xref table
//4.  depending on version the xref table may or may not be in a compressed
//    stream. If it's in a compressed stream (PDF ver 1.5+) then getting the
//    page number requires a LOT of code which is too convoluted to summarise
//    here. Otherwise it still requires a moderate amount of code ...
//5.  parse the xref table and fill a list with object numbers and offsets
//6.  handle subsections within xref table.
//7.  read 'trailer' section at end of each xref
//8.  store 'Root' object number if found in 'trailer'
//9.  if 'Prev' xref found in 'trailer' - loop back to step 3
//10. locate Root in the object list
//11. locate 'Pages' object from Root
//12. get Count from Pages.

  function GetNumber(out num: integer): boolean;
  var
    tmpStr: string;
  begin
    tmpStr := '';
    while p^ < #33 do inc(p); //skip leading CR,LF & SPC
    while (p^ in ['0'..'9']) do
    begin
      tmpStr := tmpStr + p^;
      inc(p);
    end;
    result := tmpStr <> '';
    if not result then exit;
    num := strtoint(tmpStr);
  end;

  function IsString(const str: string): boolean;
  var
    len: integer;
  begin
    len := length(str);
    result := CompareMem( p, PByte(str), len);
    if result then inc(p, len); //02-Feb-09: modified
  end;

  function FindStrInDict(const str: string): boolean;
  var
    nestLvl: integer;
    str1: AnsiChar;
  begin
    //nb: PDF 'dictionaries' start with '<<' and terminate with '>>'
    result := false;
    nestLvl := 0;
    str1 := str[1];
    while not result do
    begin
      while not (p^ in ['>','<',str1]) do inc(p);
      if (p^ = '<') then
      begin
        if (p+1)^ = '<' then begin inc(nestLvl); inc(p); end;
      end
      else if (p^ = '>') then
      begin
        if (p+1)^ = '>' then
        begin
          dec(nestLvl);
          inc(p);
          if nestLvl <= 0 then exit;
        end
      end else
      begin
        result := (nestLvl < 2) and IsString(str);
        if result then exit;
      end;
      inc(p);
    end;
  end;

  function FindEndOfDict: boolean;
  var
    nestLvl: integer;
  begin
    result := true;
    nestLvl := 1;
    while result do
    begin
      while not (p^ in ['>','<']) do inc(p);
      if (p^ = '<') then
      begin
        if (p+1)^ = '<' then begin inc(nestLvl); inc(p); end;
      end
      else if (p+1)^ = '>' then
      begin
        dec(nestLvl);
        if nestLvl < 0 then
          result := false
        else if nestLvl = 0 then
        begin
          inc(p, 2);
          exit; //found end of Dictionary
        end;
        inc(p); //skips first '>'
      end;
      inc(p);
    end;
    result := false;
  end;

  procedure SkipBlankSpace;
  begin
    while (p < pEnd) and (p^ < #33) do inc(p);
  end;

  function GotoObject(objNum: integer): boolean;
  var
    i: integer;
  begin
    i := 0;
    while i < PdfObjList.Count do
      if PPdfObj(PdfObjList[i]).number = objNum then break else inc(i);
    result :=  (i < PdfObjList.Count) and
      (PPdfObj(PdfObjList[i]).filePtr <> nil);
    if not result then exit;
    p := PPdfObj(PdfObjList[i]).filePtr;
    result := GetNumber(i) and (i = objNum);
  end;

  function IsCompressedObject(objNum: integer): boolean;
  var
    i: integer;
  begin
    i := 0;
    while i < PdfObjList.Count do
      if PPdfObj(PdfObjList[i]).number = objNum then break else inc(i);
    result :=  (i < PdfObjList.Count) and
      (PPdfObj(PdfObjList[i]).filePtr = nil);
  end;

  function GetCompressedObjectBuffer(objNum: integer;
    out buffer: PByte; out bufSize: integer): boolean;
  var
    i, N, First, objIdx: integer;
  begin
    result := false;
    i := 0;
    while i < PdfObjList.Count do
      if PPdfObj(PdfObjList[i]).number = objNum then break else inc(i);
    if (i = PdfObjList.Count) or
      (PPdfObj(PdfObjList[i]).filePtr <> nil) then exit;
    if not GotoObject(PPdfObj(PdfObjList[i]).stmObjNum) then exit;
    objIdx := PPdfObj(PdfObjList[i]).offset;

    pSaved := p;
    if not FindStrInDict('/Type') then exit;
    SkipBlankSpace;
    if not IsString('/ObjStm') then exit;

    p := pSaved;
    //todo - handle uncompressed streams too
    //assume all streams are compressed (though this is really optional)
    if not FindStrInDict('/Filter') then exit;
    SkipBlankSpace;
    //check that this a compression type that we can handle ...
    if not IsString('/FlateDecode') then exit;
    p := pSaved;
    if not FindStrInDict('/DecodeParms') or not
      FindStrInDict('/Columns') or not GetNumber(filterColCnt) then
        filterColCnt := 0; //j = column count (bytes per row)
    if filterColCnt > 0 then
    begin
      SkipBlankSpace;
      if not IsString('/Predictor') or not GetNumber(predictor) then
        predictor := 0;
    end;

    p := pSaved;
    if not FindStrInDict('/N') then exit;
    //N = number of compressed objects in stream ...
    if not GetNumber(N) or (objIdx > N) then exit;

    p := pSaved;
    if not FindStrInDict('/First') then exit;
    if not GetNumber(First) then exit; //First = first object in stream

    p := pSaved;
    if not FindStrInDict('/Length') then exit;
    if not GetNumber(k) then exit; //k = length of compressed stream
    if not FindEndOfDict then exit;
    while (p^ <> 's') do inc(p);
    if not IsString('stream') then exit;
    SkipBlankSpace;

    //decompress the stream ...
    buffer := nil;
    zlib.DecompressBuf(p, k, k*3, pointer(buffer), bufSize);
    if (filterColCnt > 0) and (predictor > 9) then
      bufSize := ReversePngFilter(buffer, filterColCnt, bufSize);

    p := buffer;
    for i := 0 to objIdx -1 do
    begin
      if not GetNumber(k) then exit;
      if not GetNumber(k) then exit;
    end;
    if not GetNumber(k) or (k <> objNum) then exit; //check it's the right obj
    if not GetNumber(k) then exit; //k = offset of obj from first obj offset

    if objIdx = N then //ie if the last object in the stream
    begin
      p := buffer + First + k;
      bufSize := buffer + bufSize - p;
    end else
    begin
      if not GetNumber(i) then exit; //skip the next object's number
      if not GetNumber(i) then exit; //i == the next object's offset
      p := buffer + First + k;
      bufSize := (i - k); //ie next object's offset minus this object's offset
    end;
    move(p^, buffer^, bufSize);
    ReallocMem(buffer, bufSize);
    p := buffer; //nb: buffer now contains just the uncompressed object
    result := bufSize > 0;
  end;

  function Find(const str: string; search_forwards: boolean = true): boolean;
  var
    len: integer;
    c: AnsiChar;
  begin
    result := false;
    len := length(str);
    if search_forwards then
    begin
      c := str[1];
      repeat
        while (p < pEnd)  and (p^ <> c) do inc(p);
        if (p = pEnd) then exit;
        if StrLComp(p, PByte(str), len) = 0 then break;
        inc(p);
      until false;
      inc(p,len);
    end else
    begin
      c := str[len];
      repeat
        while (p > PByte(ms.Memory))  and (p^ <> c) do dec(p);
        if (p = PByte(ms.Memory)) then exit;
        if StrLComp( (p-len+1), PByte(str), len) = 0 then break;
        dec(p);
      until false;
      inc(p);
    end;
    result := true;
  end;

  function PByteToInteger(buffer: PByte; byteCnt: integer): integer;
  begin
    case byteCnt of
      1: result := ord(buffer^);
      2: result := ord(buffer^) shl 8 + ord((buffer+1)^);
      3: result := ord(buffer^) shl 16 +
        ord((buffer+1)^) shl 8 + ord((buffer+2)^);
      4: result := ord(buffer^) shl 24 +
        ord((buffer+1)^) shl 16 + ord((buffer+2)^) +
        ord((buffer+3)^);
      else result := 0;
    end;
  end;

  function GetLinearizedPageNum(out pageNum: integer): boolean;
  var
    pStart,pStop: PByte;
  begin
    pageNum := -1;
    result := false;
    pStop := p + 32;
    while (p < pStop) and (p^ <> 'o') do inc(p);
    if StrLComp( p, 'obj', 3) <> 0 then exit;
    pStart := p;
    if not FindStrInDict('/Linearized') then exit;
    p := pStart;
    if FindStrInDict('/N ') and GetNumber(pageNum) then result := true;
  end;

begin
  //on error return -1 as page count
  result := -1;
  try
    ms := TMemoryStream.Create;
    PdfObjList := TList.Create;
    screen.Cursor := crHourGlass;
    application.ProcessMessages;
    try
      ms.LoadFromFile(filename);

      p := PByte(ms.Memory);
      pEnd := PByte(ms.Memory) + ms.Size;

      //for an easy life let's hope the file has a 'linearization dictionary'
      //at the beginning of the document ...
      if GetLinearizedPageNum(result) then exit;

      //find 'startxref' ignoring '%%EOF' at end of file
      p := pEnd -5;
      if not Find('startxref', false) then exit;

      rootNum := -1; //ie flag as not yet found

      if not GetNumber(k) then exit;  //xref offset ==> k
      p :=  PByte(ms.Memory) + k;

      if StrLComp(p, 'xref', 4) <> 0 then
      begin
        //most probably a cross-reference stream (ie PDF doc ver 1.5+)
        //so LOTS of hard work to do now ...

        if not GetNumber(k) then exit; //stream obj number
        if not GetNumber(k) then exit; //stream obj revision number

        pSaved := p;
        if not FindStrInDict('/Type') then exit;
        SkipBlankSpace;
        if not IsString('/XRef') then exit;

        //todo - check for and manage /Prev too

        p := pSaved;
        if not FindStrInDict('/Root') then exit;
        SkipBlankSpace;
        if not GetNumber(rootNum) then exit;
        p := pSaved;

        //todo - handle uncompressed streams too

        //assume all streams are compressed (though this is really optional)
        if not FindStrInDict('/Filter') then exit;
        SkipBlankSpace;
        //check that this a compression type that we can handle ...
        if not IsString('/FlateDecode') then exit;
        p := pSaved;
        if not FindStrInDict('/DecodeParms') or not
          FindStrInDict('/Columns') or not GetNumber(filterColCnt) then
            filterColCnt := 0; //j = column count (bytes per row)
        if filterColCnt > 0 then
        begin
          SkipBlankSpace;
          if not IsString('/Predictor') or not GetNumber(predictor) then
            predictor := 0; //Filtering used prior to Flate (PNG filter >= 10)
        end;

        //get the stream cross-ref table field sizes ...
        p := pSaved;
        if not FindStrInDict('/W') then exit;
        SkipBlankSpace;
        if p^ <> '[' then exit;
        inc(p);
        if not GetNumber(w1) or (w1 <> 1) or not GetNumber(w2) or
          not GetNumber(w3) then exit;

        //Index [F1 N1, ..., Fn, Nn]. If absent assumes F1 = 0 & N based on size
        //(Fn: first object in table subsection; Nn: number in table subsection)
        indexArray := nil;
        p := pSaved;
        if FindStrInDict('/Index') then
        begin
          SkipBlankSpace;
          if p^ <> '[' then exit;
          inc(p);
          while true do
          begin
            k := length(indexArray);
            SetLength(indexArray, k +2);
            if not GetNumber(indexArray[k]) or
              not GetNumber(indexArray[k+1]) then exit;
            pTmp2 := p;
            if not GetNumber(i) then break;
            p := pTmp2;
          end;
        end;

        p := pSaved;
        if not FindStrInDict('/Length') then exit;
        if not GetNumber(k) then exit; //k = length of compressed stream
        if not FindEndOfDict then exit;
        while (p^ <> 's') do inc(p);
        if not IsString('stream') then exit;
        SkipBlankSpace;

        //decompress the stream ...
        buffer := nil;
        try
          zlib.DecompressBuf(p, k, k*3, pointer(buffer), cnt);

          //now de-filter the decompressed output (typically PNG filtering)
          //Filter Columns should match the byte count of /W[X Y Z]
          //The decompressed stream prefiltered size == (X Y Z +1) * entries
          //(ie allowing extra byte at the start of each column for filter type)
          //see also http://www.w3.org/TR/PNG-Filters.html

          if (filterColCnt > 0) and (predictor > 9) then
            cnt := ReversePngFilter(buffer, filterColCnt, cnt);

          //make sure the decompressed & defiltered table is a valid size ...
          if cnt mod (w1 + w2 + w3) <> 0 then exit;

          //if the Index array is empty then use the default values ...
          if length(indexArray) = 0 then
          begin
            setLength(indexArray, 2);
            indexArray[0] := 0;
            indexArray[1] := cnt div (w1 + w2 + w3);
          end;

          buffPtr := buffer;
          //loop through each subsection in the table and populate our
          //object list ...
          for i := 1 to (length(indexArray) div 2) do
          begin
            k := indexArray[i*2 -2]; //k := base object number
            for j := 1 to indexArray[i*2 -1] do
            begin
              case buffPtr^ of
                #0: ;//ignore (free object)

                #1: //uncompressed object
                  begin
                    inc(buffPtr);
                    new(PdfObj);
                    PdfObjList.Add(PdfObj);
                    PdfObj.number := k;
                    PdfObj.stmObjNum := -1;
                    PdfObj.offset := PByteToInteger(buffPtr, w2);
                    PdfObj.filePtr := PByte(ms.Memory) + PdfObj.offset;
                    dec(buffPtr, w1);
                  end;
                #2: //compressed object
                  begin
                    inc(buffPtr, w1);
                    new(PdfObj);
                    PdfObjList.Add(PdfObj);
                    PdfObj.number := k;
                    PdfObj.stmObjNum := PByteToInteger(buffPtr, w2);
                    inc(buffPtr,w2);
                    PdfObj.offset := PByteToInteger(buffPtr, w3);
                    PdfObj.filePtr := nil;
                    dec(buffPtr, w1+w2);
                  end;
              end;
              inc(buffPtr, w1 + w2 + w3);
              inc(k);
            end;
          end;
        finally
          if assigned(buffer) then FreeMem(buffer);
        end;

        if rootNum < 0 then exit;
        if not GotoObject(rootNum) then exit;

        if not FindStrInDict('/Pages') then exit;
        //get the Pages' object number, go to it and get the page count ...
        if not GetNumber(pagesNum) then exit;
        k := -1;
        if IsCompressedObject(pagesNum) then
        begin
          if not GetCompressedObjectBuffer(pagesNum, buffer, cnt) then exit;
          try
            if not FindStrInDict('/Count') or not GetNumber(k) then exit;
          finally
            FreeMem(buffer);
          end;
        end
        else if not GotoObject(pagesNum) or
          not FindStrInDict('/Count') or not GetNumber(k) then exit;

        //if we get this far the page number has been FOUND!!!
        result := k;
        exit;
      end;

      //OK, most likely a PDF doc ver 1.4 (or earlier) ...

      inc(p,4); //ie skip over 'xref'

      while true do //top of loop  //////////////////////////////
      begin
        //get base object number ==> k
        if not GetNumber(k) then exit;
        //get object count ==> cnt
        if not GetNumber(cnt) then exit;
        //it is possible to have 0 objects in a section
        SkipBlankSpace;

        //add all objects in section to list ...
        for cnt := 0 to cnt-1 do
        begin
          new(PdfObj);
          PdfObjList.Add(PdfObj);
          PdfObj.number := k + cnt;
          if not GetNumber(PdfObj.offset) then exit;
          PdfObj.filePtr := PByte(ms.Memory) + PdfObj.offset;
          //while each entry SHOULD be exactly 20 bytes, not all PDF document
          //creators seems to adhere to this :( ...
          while not (p^ in [#10,#13]) do inc(p);
          while (p^ in [#10,#13]) do inc(p);
        end;
        //check for and process further subsections ...
        if p^ in ['0'..'9'] then continue;

        // parse 'trailer dictionary' ...
        if not IsString('trailer') then exit;
        pSaved := p;
        // get Root (aka Catalog) ...
        if (rootNum = -1) and FindStrInDict('/Root') then
          if not GetNumber(rootNum) then exit;
        p := pSaved;
        if not FindStrInDict('/Prev') then break; //no more xrefs

        //next xref offset ==> k
        if not GetNumber(k) then exit;
        p :=  PByte(ms.Memory) + k +4;

      end; //bottom of loop /////////////////////////////////////

      //Make sure we've got Root's object number ...
      if rootNum < 0 then exit;
      if not GotoObject(rootNum) then exit;

      if not FindStrInDict('/Pages') then exit;
      //get Pages object number then goto pagesNum ...
      if not GetNumber(pagesNum) or not GotoObject(pagesNum) then exit;
      if not FindStrInDict('/Count') then exit;
      if not GetNumber(cnt) then exit;
      //occasionally the 'count' value is an indirect object
      if GetNumber(k) and IsString(' R') then
      begin
        if not GotoObject(cnt) then exit;
        if not GetNumber(k) or //skip the generation num
          not IsString(' obj') or
          not GetNumber(cnt) then exit;
      end;
      result := cnt; //FOUND!!!!!!

    finally
      screen.Cursor := crDefault;
      for k := 0 to PdfObjList.Count -1 do
        dispose(PPdfObj(PdfObjList[k]));
      PdfObjList.Free;
      ms.Free;
    end;
  except
    //nb: errors are flagged by returning -1
  end;
end;
//------------------------------------------------------------------------------

function GetTempDirectory: string;
var
  tempFolder: array[0..MAX_PATH] of char;
begin
  if GetTempPath(MAX_PATH, @tempFolder) = 0 then
    raise Exception.Create('GetTempPath: Invalid temp path')
  else
    result := tempFolder;
end;

//------------------------------------------------------------------------------
// TMainForm methods
//------------------------------------------------------------------------------

procedure TMainForm.FormCreate(Sender: TObject);
begin
  fFileList := TStringList.Create;
  statusbar1.SimpleText := VersionInfo;
  path := extractFilePath(paramstr(0));
  tmpPath := GetTempDirectory + 'pdftkb\';
  fRotatePageCnt := -1;
  //hook the StringGrid message handler to catch WM_DROPFILES ...
  SgOldWindowProc := StringGrid1.WindowProc;
  StringGrid1.WindowProc := SgNewWindowProc;
  DragAcceptFiles(StringGrid1.Handle, True);
end;
//------------------------------------------------------------------------------

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  ClearFileList;
  fFileList.Free;
  DeleteAllTmpPdfFiles;
  RemoveDir(tmpPath);
  DragAcceptFiles(StringGrid1.Handle, False);
  StringGrid1.WindowProc := SgOldWindowProc;
end;
//------------------------------------------------------------------------------

procedure TMainForm.SgNewWindowProc(var Message: TMessage);
var
   i, cnt: integer;
   filename : array [0..MAX_PATH] of char;
  SourceFileRec: PSourceFileRec;
begin
  //handle drag and drop of pdf files onto the StringGrid control ...
  if Message.Msg = WM_DROPFILES then
  begin
    cnt := dragqueryfile(message.WParam, $FFFFFFFF, filename, MAX_PATH);
    for i := 0 to cnt-1 do
    begin
      dragqueryfile (message.WParam, i, filename, MAX_PATH);
      if fileexists(filename) and SameText(ExtractFileExt(filename),'.pdf') then
      begin
        new(SourceFileRec);
        SourceFileRec.FullPath := filename;
        SourceFileRec.Filename := ExtractFilename(filename);
        SourceFileRec.PageCount := PAGES_NOT_COUNTED;
        fFileList.AddObject(SourceFileRec.FullPath, pointer(SourceFileRec));
      end;
    end;
    dragfinish(message.WParam);
    RefreshGrid(StringGrid1.Selection.Top);
    UpdateButtonsAndMenus(StringGrid1.Selection.Top);
    tsCollateResize(nil); //in case the vertical scrollbar appears
    Message.Result := 0;
  end
  else SgOldWindowProc(Message);
end;
//------------------------------------------------------------------------------

procedure TMainForm.bCloseClick(Sender: TObject);
begin
  close;
end;
//------------------------------------------------------------------------------

procedure TMainForm.bAddClick(Sender: TObject);
var
  i: integer;
  SourceFileRec: PSourceFileRec;
  sl: TStringlist;
begin
  OpenDialog1.Options := OpenDialog1.Options + [ofAllowMultiSelect];
  if not OpenDialog1.Execute then exit;

  //we can't sort OpenDialog1.Files copy the filenames to a TStringlist ...
  sl := TStringlist.Create;
  try
    sl.Assign(OpenDialog1.Files);
    sl.sort;
    for i := 0 to sl.Count -1 do
    begin
      new(SourceFileRec);
      SourceFileRec.FullPath := sl[i];
      SourceFileRec.Filename := ExtractFilename(sl[i]);
      SourceFileRec.PageCount := PAGES_NOT_COUNTED;
      fFileList.AddObject(SourceFileRec.FullPath, pointer(SourceFileRec));
    end;
  finally
    sl.Free;
  end;
  RefreshGrid(StringGrid1.Selection.Top);
  UpdateButtonsAndMenus(StringGrid1.Selection.Top);
  tsCollateResize(nil); //in case vert scrollbar appears
end;
//------------------------------------------------------------------------------

procedure TMainForm.btnSortClick(Sender: TObject);
begin
  fFileList.Sort;
  RefreshGrid(StringGrid1.Selection.Top);
end;
//------------------------------------------------------------------------------

procedure TMainForm.RefreshGrid(RowSelected: integer);
var
  i: integer;
  s, s2: string;
begin
  with StringGrid1 do
    if fFileList.Count = 0 then
    begin
      RowCount := 1;
      Cells[0,0] := '';
      lFilename.caption := '';
      ePages.Text := '';
    end else
    begin
      RowCount := fFileList.Count;
      for i := 0 to fFileList.Count -1 do
        Cells[0,i] := PSourceFileRec(fFileList.Objects[i]).FullPath;
      with PSourceFileRec(fFileList.Objects[RowSelected])^ do
      begin
        ePages.Text := Pages;
        //get pageCount if not already done so ...
        if PageCount = PAGES_NOT_COUNTED then
          PageCount := GetPdfPageCount(FullPath);

        //truncate long filenames...
        if PageCount < 1 then
          s := format('%s  (? pages)',[Filename])
        else if PageCount = 1 then
          s := format('%s  (1 page)',[Filename])
        else
          s := format('%s  (%d pages)',[Filename, PageCount]);
        s2 := s;
        i := self.Canvas.TextWidth(s);
        while i > 228 do
        begin
          delete(s,8,1);
          i := self.Canvas.TextWidth(s);
        end;
        if s <> s2 then insert('...', s, 8);
        lFilename.caption := s;
      end;
    end;
end;
//------------------------------------------------------------------------------

procedure TMainForm.UpdateButtonsAndMenus(RowSelected: integer);
begin
  if PageControl1.ActivePage = tsCollate then
  begin
    if fFileList.Count = 0 then
    begin
      btnBuild.Enabled := false;
      bDelete.Enabled := false;
      bMoveUp.Enabled := false;
      bMoveDown.Enabled := false;
      ePages.Enabled := false;
      lPages.Enabled := false;
      ePages.Enabled := false;
    end else
    begin
      btnBuild.Enabled := true;
      bDelete.Enabled := true;
      bMoveUp.Enabled := RowSelected > 0;
      bMoveDown.Enabled := RowSelected < StringGrid1.RowCount -1;
      ePages.Enabled := (fFileList.Count < 27);
      lPages.Enabled := ePages.Enabled;
    end;
  end
  else if PageControl1.ActivePage = tsSplit then
  begin
    btnBuild.Enabled := fileExists(lblSplitSrc.Caption);
  end
  else if PageControl1.ActivePage = tsBackground then
  begin
    btnBuild.Enabled := fileExists(lblSource_background.Caption) and
      fileExists(lblBackground.Caption);
  end;
  statusbar1.SimpleText := VersionInfo;
end;
//------------------------------------------------------------------------------

procedure TMainForm.PageControl1Change(Sender: TObject);
begin
  UpdateButtonsAndMenus(StringGrid1.Selection.Top);
  if PageControl1.ActivePage = tsSplit then
    btnBuild.Caption := 'S&ave' else
    btnBuild.Caption := 'Save &As';
end;
//------------------------------------------------------------------------------

procedure TMainForm.ePagesChange(Sender: TObject);
var
  i: integer;
  isValidPageRange: boolean;
begin
  if not ePages.Focused then exit;
  btnBuild.Enabled := true;
  i := StringGrid1.Selection.Top;
  if (i >= fFileList.Count) then exit;
  PSourceFileRec(fFileList.Objects[i]).Pages := ePages.Text;
  RefreshGrid(i);
  FmtPagesString(ePages.Text, isValidPageRange);
  if isValidPageRange then
    ePages.Font.Color := clBlack else
    ePages.Font.Color := clRed;
end;
//------------------------------------------------------------------------------


function TMainForm.FmtPagesString(PagesStr: string; out IsValid: boolean): string;
var
  i, len: integer;
  rngeStrt, rngeEnd: string;

  procedure SkipSpaces;
  begin
    while (i <= len) and (PagesStr[i] = ' ') do inc(i);
  end;

  function GetNumber(var num: string): boolean;
  begin
    num := '';
    while (i <= len) and (PagesStr[i] in ['0'..'9']) do
    begin
      num := num + PagesStr[i];
      inc(i);
    end;
    result := num <> '';
  end;

begin
  //returns 1 or more page ranges, each separated by spc and finishing with spc
  //eg - Input: "1, 2-3,19, 21 - 28" Output: "1 3-6 19 21-28 "
  IsValid := false;
  result := '';
  len := length(PagesStr);
  i := 1;
  while i <= len do
  begin
    SkipSpaces;
    if (i > len) then break;
    if not GetNumber(rngeStrt) then exit;
    SkipSpaces;
    if (i <= len) and (PagesStr[i] = '-') then
    begin
      inc(i);
      SkipSpaces;
      if not GetNumber(rngeEnd) then exit;
      SkipSpaces;
      result := result + rngeStrt + '-' + rngeEnd + ' ';
    end
    else
      result := result + rngeStrt + ' ';

    if (i > len) then break
    else if (PagesStr[i] in [',',';',':']) then
    begin
      inc(i);
      continue;
    end else
      exit;
  end;
  if result = '' then result := ' ';
  IsValid := true;
end;
//------------------------------------------------------------------------------

procedure TMainForm.ClearFileList;
var
  i: integer;
begin
  for i := 0 to fFileList.Count -1 do
    dispose(PSourceFileRec(fFileList.Objects[i]));
  fFileList.Clear;
end;
//------------------------------------------------------------------------------

procedure TMainForm.StringGrid1SelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
  UpdateButtonsAndMenus(ARow);
  RefreshGrid(ARow);
end;
//------------------------------------------------------------------------------

procedure TMainForm.bMoveUpClick(Sender: TObject);
var
  i: integer;
  sel: TGridRect;
  SourceFileRec: PSourceFileRec;
begin
  sel := StringGrid1.Selection;
  i := sel.Top;
  SourceFileRec := pointer(fFileList.Objects[i]);
  fFileList.Delete(i);
  fFileList.InsertObject(i-1,SourceFileRec.FullPath,pointer(SourceFileRec));
  dec(sel.Top);
  dec(sel.Bottom);
  RefreshGrid(i-1);
  StringGrid1.Selection := sel;
  if sel.Top < StringGrid1.TopRow then StringGrid1.TopRow := sel.Top;
  UpdateButtonsAndMenus(i-1);
end;
//------------------------------------------------------------------------------

procedure TMainForm.bMoveDownClick(Sender: TObject);
var
  i: integer;
  sel: TGridRect;
  SourceFileRec: PSourceFileRec;
begin
  sel := StringGrid1.Selection;
  i := sel.Top;
  SourceFileRec := pointer(fFileList.Objects[i]);
  fFileList.Delete(i);
  fFileList.InsertObject(i+1,SourceFileRec.FullPath,pointer(SourceFileRec));
  inc(sel.Top);
  inc(sel.Bottom);
  RefreshGrid(i+1);
  StringGrid1.Selection := sel;
  with StringGrid1 do
    if sel.Bottom  >= TopRow + VisibleRowCount then
      StringGrid1.TopRow := sel.Bottom - VisibleRowCount +1;
  UpdateButtonsAndMenus(i+1);
end;
//------------------------------------------------------------------------------

procedure TMainForm.bDeleteClick(Sender: TObject);
var
  i: integer;
begin
  i := StringGrid1.Selection.Top;
  dispose(PSourceFileRec(fFileList.Objects[i]));
  fFileList.Delete(i);
  if i = fFileList.Count then dec(i);
  RefreshGrid(i);
  UpdateButtonsAndMenus(i);
end;
//------------------------------------------------------------------------------

procedure TMainForm.DeleteAllTmpPdfFiles;
var
  i: integer;
  sr: TSearchRec;
begin
  i := FindFirst(tmpPath+'*.pdf', faAnyFile, sr);
  while i = 0 do
  begin
    SetFileAttributes(pchar(tmpPath+sr.Name), 0); //remove read-only etc
    DeleteFile(tmpPath+sr.Name);
    i := FindNext(sr);
  end;
  FindClose(sr);
end;
//------------------------------------------------------------------------------

procedure TMainForm.btnBuildClick(Sender: TObject);
var
  j: integer;
  i: integer;
  s, action, fileParams, outfileParams, actionParams,
    allowParams, ownerParam, userParam: string;
  IsValidPageRanges: boolean;
  succeeded: boolean;

  function ExecuteCommand(const command: string): boolean;
  var
    resultStr: string;
  begin
    //showmessage(command); //debugging only
    screen.Cursor := crHourGlass;
    application.ProcessMessages;
    try
      ErrorNum := WinExecAndWait32(command, SW_HIDE, 0, resultStr);
      result := ErrorNum = 0;
      if not result then ErrorMessage := resultStr;
    finally
      screen.Cursor := crDefault;
    end;
  end;

begin
  ErrorMessage := '';
  ErrorNum := 0;
  
  btnBuild.Enabled := false;
  try

    if not FileExists(path+'pdftk.exe') then
    begin
      application.MessageBox('pdftk.exe does not exist in the application folder.',
        pchar(application.Title), MB_ICONSTOP);
      exit;
    end;

    if (eUserPassword.text = '') then
      userParam := '' else
      userParam := 'user_pw ' + eUserPassword.text;

    if (eOwnerPassword.text = '') then
      ownerParam := '' else
      ownerParam := 'owner_pw ' + eOwnerPassword.text;

    allowParams := '';
    if (ownerParam <> '') or (userParam <> '') then
    begin
      if cbAllowAll.Checked then
        allowParams := 'allow allfeatures '
      else
      begin
        if cbPrinting.Checked then allowParams := allowParams + 'printing ';
        if cbDegradedPrinting.Checked then allowParams := allowParams + 'degradedprinting ';
        if cbModifyContents.Checked then allowParams := allowParams +'modifycontents ';
        if cbAssembly.Checked then allowParams := allowParams +'assembly ';
        if cbCopyContents.Checked then allowParams := allowParams +'copycontents ';
        if cbScreenReaders.Checked then allowParams := allowParams +'screenreaders ';
        if cbModifyAnnotations.Checked then allowParams := allowParams +'modifyannotations ';
        if cbFillIn.Checked then allowParams := allowParams +'fillin ';
        if allowParams <> '' then allowParams := 'allow ' + allowParams;
      end;
    end;


    //............................................................................
    //COLLATE PDF DOCUMENTS
    //............................................................................
    if PageControl1.ActivePage = tsCollate then
    begin
      if (fFileList.Count = 0) or not SaveDialog1.Execute then exit;

      //first, make sure we're not saving over a source file ...
      for i := 0 to fFileList.Count -1 do
        if PSourceFileRec(fFileList.Objects[i]).FullPath = SaveDialog1.FileName then
        begin
          beep;
          statusbar1.SimpleText := '  Error - can''t save over a source file.';
          exit;
        end;

      //nb: Since PDFTK has limited 'file handles' (ie A..Z), if there are
      //more than 26 files, we simply collate the files using the '*' wildcard
      //char. (However, using '*' prevents page selection for individual files.)
      if fFileList.Count > 26 then
      begin
        ForceDirectories(tmpPath);
        DeleteAllTmpPdfFiles;

        //copy all the pdf files into a temp folder ready for collating
        //while preserving sort order ...
        for i := 0 to fFileList.Count -1 do
          with PSourceFileRec(fFileList.Objects[i])^ do
            CopyFile(
              pchar(FullPath),
              pchar(format('%s%3.3d.pdf',[tmpPath, i+1])),
              false);

        try
          ChDir(tmpPath); //workaround as pdftk doesn't like a path before *.pdf!
                          //not sure how Vista friendly this is but should be OK.
          s := format('"%spdftk.exe" *.pdf cat output "%s" %s %s %s dont_ask',
            [path, SaveDialog1.FileName, allowParams, ownerParam, userParam]);
          succeeded := executeCommand(s);
        finally
          DeleteAllTmpPdfFiles;
        end;
      end else
      begin
        //ie less than 26 files ...
        actionParams := '';                //just stops warnings
        fileParams := '';
        for i := 0 to fFileList.Count -1 do
          with PSourceFileRec(fFileList.Objects[i])^ do
          begin
            fileParams :=
              format('%s %s="%s"',[fileParams, chr(ord('A')+ i), FullPath]);

            actionParams := actionParams + chr(ord('A')+i) +
              FmtPagesString(Pages, IsValidPageRanges);

            if not IsValidPageRanges then
            begin
              beep;
              statusbar1.SimpleText := '  Invalid page range specified';
              exit;
            end;
          end;

        s := format('"%spdftk.exe" %s cat %s output "%s" %s %s %s dont_ask',
          [path, fileParams, actionParams, SaveDialog1.FileName, allowParams,
          ownerParam, userParam]);
        succeeded := executeCommand(s);
      end;
    end
    //............................................................................
    //SPLIT PDF DOCUMENTS
    //............................................................................
    else if PageControl1.ActivePage = tsSplit then
    begin
      outfileParams := changeFileExt(lblSplitSrc.Caption,'') + '_%03d.pdf';
      succeeded := executeCommand(format('%spdftk.exe A="%s" burst output "%s" %s %s %s dont_ask',
        [path, lblSplitSrc.Caption, outfileParams, allowParams, ownerParam, userParam]));
    end
    //............................................................................
    //BACKGROUND/STAMP PDF DOCUMENTS
    //............................................................................
    else if PageControl1.ActivePage = tsBackground then
    begin
      if not fileExists(lblSource_background.Caption) or
        not SaveDialog1.Execute then exit;

      if lblSource_background.Caption = SaveDialog1.FileName then
      begin
        beep;
        statusbar1.SimpleText := '  Error - can''t save over a source file.';
        exit;
      end;

      if rbBackground.Checked then
        action := 'background' else
        action := 'stamp';

      if cbPageOneOnly.Checked and (GetPdfPageCount(lblSource_background.Caption) > 1) then
      begin
        ForceDirectories(tmpPath);
        DeleteAllTmpPdfFiles;

        //split off page 1 to tmp1.pdf ...
        s := format('"%spdftk.exe" A="%s" cat A1 output "%stmp1.pdf" %s %s %s dont_ask',
          [path, lblSource_background.Caption, tmpPath,
          allowParams, ownerParam, userParam]);
        succeeded := executeCommand(s);
        //create tmp3.pdf from backgrounded page 1 ...
        succeeded := succeeded and
          executeCommand(format('%spdftk.exe A="%stmp1.pdf" %s "%s" output "%stmp3.pdf" %s %s %s dont_ask',
          [path, tmpPath, action, lblBackground.Caption,
          tmpPath, allowParams, ownerParam, userParam]));

        //split off page 2 to end to tmp2.pdf ...
        s := format('"%spdftk.exe" A="%s" cat A2-end output "%stmp2.pdf" %s %s %s dont_ask',
          [path, lblSource_background.Caption, tmpPath,
          allowParams, ownerParam, userParam]);
        succeeded := succeeded and executeCommand(s);
        //join the result back into one pdf document ...
        s := format('"%spdftk.exe" A="%stmp3.pdf" B="%stmp2.pdf" cat A B output "%s" %s %s %s dont_ask',
          [path, tmpPath, tmpPath, SaveDialog1.FileName, allowParams, ownerParam, userParam]);
        succeeded := succeeded and executeCommand(s);
        DeleteAllTmpPdfFiles;
      end else
      begin
        succeeded := executeCommand(format('%spdftk.exe A="%s" %s "%s" output "%s" %s %s %s dont_ask',
          [path, lblSource_background.Caption, action, lblBackground.Caption,
          SaveDialog1.FileName, allowParams, ownerParam, userParam]));
      end;
    end
    //............................................................................
    //ROTATE PDF DOCUMENTS
    //............................................................................
    else if PageControl1.ActivePage = tsRotate then
    begin

      if not SaveDialog1.Execute then exit;

      //first, make sure we're not saving over a source file ...
      if lblSource_rotate.Caption = SaveDialog1.FileName then
        begin
          beep;
          statusbar1.SimpleText := '  Error - can''t save over a source file.';
          exit;
        end;

      if rbRotRight.Checked then
        action := 'R'
      else if rbRotLeft.Checked then
        action := 'L'
      else
        action := 'D';

      i := strtoint(eRotPagesFrom.Text);
      j := strtoint(eRotPagesTo.Text);
      if j < i then j := i;

      if i < 0 then
        exit //should never happen
      else if i = 1 then
        actionParams := ''
      else if i = 2 then
        actionParams := '1 '
      else
        actionParams := format('1-%d ', [i-1]);

      if j > i then
        actionParams := actionParams + format('%d-%d%s', [i,j, action]) else
        actionParams := actionParams + format('%d%s', [i,action]);

      if j = fRotatePageCnt then
        //do nothing
      else if j = fRotatePageCnt -1 then
        actionParams := actionParams + format(' %d', [fRotatePageCnt])
      else
        actionParams := actionParams + format(' %d-%d', [j+1, fRotatePageCnt]);

      s := format('"%spdftk.exe" "%s" cat %s output "%s" %s %s %s dont_ask',
        [path, lblSource_rotate.Caption, actionParams, SaveDialog1.FileName,
        allowParams, ownerParam, userParam]);
      succeeded := executeCommand(s);
    end
    else exit;

    //............................................................................

    if succeeded then
    begin
      statusbar1.SimpleText := '  '+ SaveDialog1.FileName + ' created.';
      bClose.SetFocus;
    end else
    begin
      application.MessageBox(pchar('Operation failed.'#10+'Error number: $'+
        inttohex(ErrorNum,8)+#10+ ErrorMessage),
        pchar(application.Title), MB_ICONSTOP);
    end;

  finally
    btnBuild.Enabled := true;
  end;

end;
//------------------------------------------------------------------------------

procedure TMainForm.cbAllowAllClick(Sender: TObject);
begin
  if not cbAllowAll.Focused then exit;
  cbPrinting.Checked := cbAllowAll.Checked;
  cbDegradedPrinting.Checked := cbAllowAll.Checked;
  cbModifyContents.Checked := cbAllowAll.Checked;
  cbAssembly.Checked := cbAllowAll.Checked;
  cbCopyContents.Checked := cbAllowAll.Checked;
  cbScreenReaders.Checked := cbAllowAll.Checked;
  cbModifyAnnotations.Checked := cbAllowAll.Checked;
  cbFillIn.Checked := cbAllowAll.Checked;
end;
//------------------------------------------------------------------------------

procedure TMainForm.cbPrintingClick(Sender: TObject);
begin
  //if any checkbox is unticked then make sure 'all actions' is unticked too.
  if (Sender is TCheckbox) and
    TCheckbox(Sender).Focused and not TCheckbox(Sender).Checked then
      cbAllowAll.Checked := false;
end;
//------------------------------------------------------------------------------

procedure TMainForm.btnSrcBackgroundClick(Sender: TObject);
begin
  OpenDialog1.Options := OpenDialog1.Options - [ofAllowMultiSelect];
  if not OpenDialog1.Execute then exit;
  lblSource_background.Caption := OpenDialog1.FileName;
  UpdateButtonsAndMenus(0);
end;
//------------------------------------------------------------------------------

procedure TMainForm.Button2Click(Sender: TObject);
begin
  OpenDialog1.Options := OpenDialog1.Options - [ofAllowMultiSelect];
  if not OpenDialog1.Execute then exit;
  lblBackground.Caption := OpenDialog1.FileName;
  UpdateButtonsAndMenus(0);
end;
//------------------------------------------------------------------------------

procedure TMainForm.About1Click(Sender: TObject);
begin
  with TAboutForm.Create(self) do
  try
    lblVersion.Caption := versionInfo;
    showModal;
  finally
    free;
  end;
end;
//------------------------------------------------------------------------------

procedure TMainForm.Contents1Click(Sender: TObject);
begin
  Application.HelpFile := path + 'PdftkBuilder.hlp';
  Application.HelpCommand(HELP_CONTENTS, 0);
end;
//------------------------------------------------------------------------------

procedure TMainForm.btnSplitSrcClick(Sender: TObject);
begin
  OpenDialog1.Options := OpenDialog1.Options - [ofAllowMultiSelect];
  if not OpenDialog1.Execute then exit;
  lblSplitSrc.Caption := OpenDialog1.FileName;
  UpdateButtonsAndMenus(0);
end;
//------------------------------------------------------------------------------

procedure TMainForm.eRotPagesFromChange(Sender: TObject);
var
  val: integer;
  edit: TEdit;
  OkVal: boolean;
begin
  edit := TEdit(Sender);
  val := StrToIntDef(edit.Text, -1);
  OkVal := (val > 0) and (val <= fRotatePageCnt);
  if OkVal then
    edit.Font.Color := clBlack else
    edit.Font.Color := clRed;
  btnBuild.Enabled := (lblSource_rotate.Caption <> '') and
    (eRotPagesFrom.Font.Color = clBlack) and (eRotPagesTo.Font.Color = clBlack);
end;
//------------------------------------------------------------------------------

procedure TMainForm.btnSrcRotateClick(Sender: TObject);
begin
  if not OpenDialog1.Execute then exit;
  lblSource_rotate.Caption := OpenDialog1.FileName;
  fRotatePageCnt := GetPdfPageCount(OpenDialog1.FileName);
  eRotPagesFromChange(eRotPagesFrom);
  eRotPagesFromChange(eRotPagesTo);
end;
//------------------------------------------------------------------------------

procedure TMainForm.tsCollateResize(Sender: TObject);
var
  i,j, w: integer;
begin
  w := StringGrid1.ClientWidth;
  with StringGrid1 do
    for i := 0 to RowCount -1 do
    begin
      j := canvas.TextWidth(Cells[0,i])+4;
      if j > w then w := j;
    end;
  StringGrid1.ColWidths[0] := w;
end;
//------------------------------------------------------------------------------

//Below is a workaround for a Windows bug ...
//when opening an Open/Save dialog more than once and hovering over a filename
//causing a hint window to popup, the application will crash on Windows XP Sp2.
//(If you're sceptical about this being a Windows bug then check it out
//in Notepad.exe.)

initialization
OleInitialize(nil);

finalization
OleUninitialize;

end.
